DESCRIPTION = "description"
ATTRIBUTES = "attributes"
CALLS = "calls"
SPEC = "spec"
MOCKS = "mocks"
PATCHES = "patches"
RESULT = "result"
EXCEPTION = "exception"
TYPE = "type"
MESSAGE = "message"
RETURN_VALUE = "return"
PARAMETERS = "parameters"
VALUE = "value"
INIT_PARAMETERS = "init_parameters"
TARGET = "target"
CLASS_NAME = "class_name"
NAME = "name"
SHORT_NAME = "short_name"
PATCH_REPLACEMENT = "replacement"
TESTS = "tests"
MODULE = "module"
FORMAT_VERSION = "format_version"
