from __future__ import annotations

INVALID_TEST_INDEX = -1
GET_TEST_NAME_PROMPT = (
    "Test name help: 'test_' prefix will be added automatically. "
    "Use a leading '_' to include the target function name as "
    "part of the prefix.\n"
    "Test name:"
)
