class SingleLineString:
    """
    A special type to represent a string type that should only contain one line
    of text.
    """

    pass


class NoDefault:
    """
    Marker object for not having a default value.
    Note that None can be a valid value.
    """
